/* =====================================================================================
                      The UQ Toolkit (UQTk) version 2.1
                     Copyright (2013) Sandia Corporation
                      http://www.sandia.gov/UQToolkit/

     Copyright (2013) Sandia Corporation. Under the terms of Contract DE-AC04-94AL85000
     with Sandia Corporation, the U.S. Government retains certain rights in this software.

     This file is part of The UQ Toolkit (UQTk)

     UQTk is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     UQTk is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU Lesser General Public License for more details.
 
     You should have received a copy of the GNU Lesser General Public License
     along with UQTk.  If not, see <http://www.gnu.org/licenses/>.
 
     Questions? Contact Bert Debusschere <bjdebus@sandia.gov>
     Sandia National Laboratories, Livermore, CA, USA
===================================================================================== */
#include <math.h>
#include "error_handlers.h"
#include "ftndefs.h"
#include "gen_defs.h"

#include "Array1D.h"
#include "Array2D.h"
#include "dsfmt_add.h"
#include "uqtkmcmc.h"

#define DBL_MAX  1.7e+308

// Computes Cholesky decomposition of a real symmetric matrix A
extern FTN_FUNC void FTN_NAME(dpotrf)(char*, int*, double*, int*, int*);

MCMC::MCMC(double (*logPosterior)(Array1D<double>&, void *),void *postinfo)
{

  /// Copy the constructor arguments to the internal variables
  postInfo_     = postinfo;
  logPosterior_ = logPosterior;

  /// Set the initialization flags to false
  chaindimInit_ = false;
  propcovInit_  = false;
  methodInit_   = false;
  outputInit_   = false;
  adaptstepInit_= false;
  gammaInit_    = false;
  epscovInit_   = false;

  /// Initiate the random number generator seed
  /// \todo This needs to be made more generic
  seed_=13;
  dsfmt_gv_init_gen_rand( seed_ );

  /// Set the location of the last chain state written (-1 means nothing is written to files yet)
  lastwrite_=-1;

  /// By default the names are not prepended
  namePrepend_=false;

  /// Set defaults
  this->initDefaults();

  return;
}

void MCMC::initDefaults()
{
  this->default_method_="am";
  this->default_gamma_=0.01;
  this->default_eps_cov_=1e-8;
  
  return;
}

void MCMC::printChainSetup()
{
  if (this->methodInit_)
    cout << "Method           : " << this->methodinfo_.type << endl;
  else 
    cout << "Method (default) : " << this->default_method_ << endl;
  if (this->gammaInit_)
    cout << "Gamma            : " << this->methodinfo_.gamma << endl;
  else 
    cout << "Gamma (default)  : " << this->default_gamma_ << endl;
  if (this->epscovInit_)
    cout << "Eps_Cov          : " << this->methodinfo_.eps_cov << endl;
  else 
    cout << "Eps_Cov (default): " << this->default_eps_cov_ << endl;
  return;
}



double MCMC::evalLogPosterior(Array1D<double>& m)
{
  /// Evaluate given the log-posterior function defined by the user in the constructor
  return logPosterior_(m,postInfo_);
}

void MCMC::initMethod(string method)
{
  /// Set the method type
  methodinfo_.type=method;
  /// Set the initialization flag to True
  methodInit_=true;
  return;
}

void MCMC::initAdaptSteps(int adaptstart,int adaptstep, int adaptend)
{
  /// Initialize the vector containing the adaptivity information: when to start, how often and when to stop.
  methodinfo_.adaptstep.Resize(3);
  methodinfo_.adaptstep(0) = adaptstart;
  methodinfo_.adaptstep(1) = adaptstep;
  methodinfo_.adaptstep(2) = adaptend;

  /// Set the initialization flag to True
  adaptstepInit_=true;

  return;
}

void MCMC::initAMGamma(double gamma)
{
  /// Initialize the scale factor gamma
  methodinfo_.gamma = gamma;
  /// Set the initialization flag to True
  gammaInit_=true;

  return;
}

void MCMC::initEpsCov(double eps_cov)
{
  /// Initialize the covariance 'nugget'
  methodinfo_.eps_cov=eps_cov;
  /// Set the initialization flag to True
  epscovInit_=true;

  return;
}

void MCMC::setOutputInfo(string outtype, string file,int freq_file, int freq_screen)
{
  outputinfo_.type=outtype;
  outputinfo_.filename=file;
  outputinfo_.freq_chainfile=freq_file;
  outputinfo_.freq_outscreen=freq_screen;
  /// Set the initialization flag to True
  outputInit_=true;
  return;
}

void MCMC::initChainPropCov(Array2D<double>& propcov)
{
  /// Initialize the proposal covariance matrix
  methodinfo_.chcov = propcov;
  /// Set the initialization flag to True
  propcovInit_ = true;
  return;
}

void MCMC::initChainPropCovDiag(Array1D<double>& sig)
{

  /// Create a diagonal matrix and fill in the diagonal terms
  methodinfo_.chcov.Resize(this->chainDim_,this->chainDim_,0.e0);
  for(int i=0;i<this->chainDim_;i++) methodinfo_.chcov(i,i)=sig(i)*sig(i);
  /// Set the initialization flag to True
  propcovInit_=true;
  return;
}

void MCMC::runChain(int ncalls, Array1D<double>& chstart)
{
  /// Check the mandatory initialization
  if(!chaindimInit_)
    throw Tantrum((string) "Chain dimensionality needs to be initialized");

  /// Check what is not initialized and use defaults instead
  /// \todo Specify defaults somewhere more transparently

  /// Set defaults proposal covariance
  if(!propcovInit_){
    Array1D<double> chsig(this->chainDim_,0.e0);
    for(int i=0;i<this->chainDim_;i++) chsig(i)=MAX(fabs(0.1*chstart(i)),0.001);
    this->initChainPropCovDiag(chsig);
  }

  /// Set defaults output format
  if (!outputInit_)
    this->setOutputInfo("txt","chain.dat", (int) ncalls/100, (int) ncalls/20);
    
  /// Set the default method
  if (!methodInit_)
    methodinfo_.type=default_method_;
  
  /// Set the default parameters for aMCMC
  if(!strcmp(this->methodinfo_.type.c_str(),"am")){ 
    
    if(!adaptstepInit_)
      this->initAdaptSteps((int) ncalls/10,10,ncalls);

    if(!gammaInit_)
      this->initAMGamma(this->default_gamma_);

    if(!epscovInit_)
      this->initEpsCov(this->default_eps_cov_);
    
  }

  /// For simplicity, work variables
  string method = methodinfo_.type;
  string output = outputinfo_.type;

  /// Set the number of substeps per one chain step
  if(!strcmp(method.c_str(),"ss"))
    nSubSteps_=chainDim_;
  else //if(!strcmp(method.c_str(),"am")) or if(!strcmp(method.c_str(),"mala"))
    nSubSteps_=1;
 
  /// Initial chain state
  currState_.step  = 0;
  currState_.state = chstart;
  currState_.alfa  = 0.0;
  currState_.post  = this->evalLogPosterior(chstart);
  this->updateMode();
  fullChain_.PushBack(currState_);
  
  /// Number of accepted steps and number of all trials
  int nacc=0;
  int nall=0;

  /// Main loop
  for (int t=1; t <= ncalls; t++) {
    currState_.step=t;
    double sum_alpha=0.0;

    /// Create a new proposed sample
    /// If the method is 'ss'(Single-Site), one actually searches all dimensions before recording the new state, i.e.
    /// one chain step has d substeps, where d is the chain dimensionality.
    /// For 'am'(Adaptive), d is set to 1.
    for (int is=0; is<nSubSteps_; is++){
      Array1D<double> m_cand;
      if(!strcmp(method.c_str(),"ss"))
	this->proposalSingleSite(currState_.state, m_cand, is);
      else if(!strcmp(method.c_str(),"am"))
      	this->proposalAdaptive(currState_.state, m_cand, t);
      else
	throw Tantrum((string) "Chain running method is not recognized");

      /// Evaluate the posterior at the new sample point
      double eval_cand = this->evalLogPosterior(m_cand);

      /// Accept or reject it
      double alpha = exp(eval_cand - currState_.post); 
      sum_alpha+=alpha;
      if (alpha>=1 || alpha > dsfmt_gv_genrand_urv()){
	if (this->inDomain(m_cand)) {
	  nacc++;
          currState_.state = m_cand;
          currState_.post = eval_cand;
	}
      }
    
      nall++;
    }
    currState_.alfa=sum_alpha/nSubSteps_;

    /// Append the current state to the array of all past states
    fullChain_.PushBack(currState_);

    /// Keep track of the mode (among the locations visited so far)
    if (currState_.post > modeState_.post)
      this->updateMode();

    /// Output to the screen
    if( t % outputinfo_.freq_outscreen == 0 || t==ncalls){
      
      printf("%lg %% completed; Chain step %d\n", 100.*t/ncalls,t); 
      printf("================= Current logpost:%f, Max logpost:%f, Accept rate:%f\n",currState_.post,modeState_.post,(double) nacc/nall); 
      printf("================= Current MAP params: ");
      for(int ic=0;ic<this->chainDim_;ic++)
	printf("par(%d)=%f ",ic,modeState_.state(ic));
      cout << endl;
      
    }

    /// Output to file
    if( t % outputinfo_.freq_chainfile == 0 || t==ncalls){
      
      if(!strcmp(output.c_str(),"txt"))
	this->writeChainTxt(outputinfo_.filename);
      else  if(!strcmp(output.c_str(),"bin"))
	this->writeChainBin(outputinfo_.filename);
      else
	throw Tantrum((string) "Chain output type is not recognized");
      
      lastwrite_ = t;
      
    }

  }  /// End of main loop  

  return;
}

void MCMC::updateMode()
{
  /// Update the chain mode (MAP state)
  modeState_.step  = currState_.step;
  modeState_.state = currState_.state;
  modeState_.post  = currState_.post;
  modeState_.alfa  = -1.0;
  
  return;
}

void MCMC::proposalSingleSite(Array1D<double>& m_t,Array1D<double>& m_cand,int dim)
{
  /// Single-site proposal
  m_cand=m_t;
  m_cand(dim) += ( sqrt(methodinfo_.chcov(dim,dim))*dsfmt_gv_genrand_nrv() );

  return;
}


bool MCMC::inDomain(Array1D<double>& m)
{
    int nd = m.XSize();
    
    for (int id=0;id<nd;id++){
        if (!lower_flag_(id))
            return true;
        else if (m(id)<Lower_(id))
            return false;

        if (!upper_flag_(id))
            return true;
        else if (m(id)>Upper_(id))
            return false;
    }
    
    return true;
}

void MCMC::setLower(double lower, int i)
{
        this->Lower_(i)=lower;
        lower_flag_(i)=1;
    
    return;
}

void MCMC::setUpper(double upper, int i)
{
    this->Upper_(i)=upper;
    upper_flag_(i)=1;
    return;
}

void MCMC::setChainDim(int chdim)
{
    this->chainDim_=chdim; chaindimInit_=true;
    this->setDefaultDomain();
    
    return;
}

void MCMC::setDefaultDomain()
{
    this->Lower_.Resize(this->chainDim_,-DBL_MAX);
    this->Upper_.Resize(this->chainDim_, DBL_MAX);
    this->lower_flag_.Resize(this->chainDim_,0);
    this->upper_flag_.Resize(this->chainDim_,0);

    
    return;
}

void MCMC::proposalAdaptive(Array1D<double>& m_t,Array1D<double>& m_cand,int t)
{
  int chol_info=0;
  char lu='L';

  if (t == 1) {
    // at the first iteration, the mean is easy and the sample covariance is 0
    methodinfo_.curmean=m_t;
    methodinfo_.curcov.Resize(this->chainDim_,this->chainDim_,0.e0);
  } else if( t > 1 && t < methodinfo_.adaptstep(2) ){
    // after the first iteration, start keeping track of the sample mean
    for (int i=0; i < chainDim_; i++) {
      methodinfo_.curmean(i)  = ( methodinfo_.curmean(i)*(t-1.) + m_t(i) )/t;
    }

    for (int i=0; i < this->chainDim_; i++) 
      for (int j=0; j < i+1; j++)
        methodinfo_.curcov(i,j) = ( (t-2.)/(t-1.) )*methodinfo_.curcov(i,j) +
	  ( t/((t-1.)*(t-1.)) ) * ( m_t(i) - methodinfo_.curmean(i) )*( m_t(j) - methodinfo_.curmean(j) );

    // transpose
    for (int i=0; i < chainDim_; i++) 
      for (int j=i+1; j < chainDim_ ; j++)
        methodinfo_.curcov(i,j) = methodinfo_.curcov(j,i) ; 

  }

  // Jump size
  double sigma = methodinfo_.gamma * 2.4 * 2.4 / (double)this->chainDim_;

  if ( t == 1 ) {
    propLCov_=methodinfo_.chcov;

    // Cholesky factorization of the proposal covariance propLCov_, done in-place 
    // Note, for diagonal covariances, this is an overkill
    FTN_NAME(dpotrf)(&lu,&chainDim_, propLCov_.GetArrayPointer(), &chainDim_, &chol_info);

  }

  if ( ( t > methodinfo_.adaptstep(0) ) &&
       ( (t % methodinfo_.adaptstep(1) ) ==  0 ) && 
       (t <= methodinfo_.adaptstep(2)  )
     ) {
      
    for (int i=0; i < chainDim_; i++)
      for (int j=0; j < chainDim_; j++)
	propLCov_(i,j) = sigma*( methodinfo_.curcov(i,j) + (i==j)*methodinfo_.eps_cov ) ;

    // Cholsky factorization of the proposal covariance propLCov_, done in-place 
    FTN_NAME(dpotrf)(&lu,&chainDim_, propLCov_.GetArrayPointer(),&chainDim_,&chol_info);

    // Catch the error in Cholesky factorization
    if (chol_info != 0 ) {
      printf("Error in Cholesky factorization, info=%d, printing the matrix below:\n", chol_info);
      for(int i=0;i<chainDim_;i++){
        for(int j=0;j<chainDim_;j++)
          printf("%lg ",propLCov_(i,j));
        printf("\n");
      }
      exit(1);
    }
  }
  

  // Candidate state is a multivariate normal sample away from the current state
  m_cand=m_t;
  for (int i=0; i < chainDim_; i++) {
    double Lnrv=0.0;
    for (int j=0; j < i+1; j++) {
      Lnrv += propLCov_(i,j)*dsfmt_gv_genrand_nrv();
    }
    m_cand(i) += Lnrv;
  }

  return;

}

void MCMC::writeChainTxt(string filename)
{
 
  /// Choose whether write or append
  char* writemode=(char *)"w";
  if (lastwrite_>=0 || namePrepend_)
    writemode=(char *)"a";
  
  /// Open the text file
  FILE* f_out;  
  if(!(f_out = fopen(filename.c_str(),writemode))){ 
    printf("writeChain: could not open file '%s'\n",filename.c_str()); 
    exit(1); 
  }

  /// Write to the text file
  for(int i=this->lastwrite_+1;i<(int) this->fullChain_.XSize();i++){
    fprintf(f_out, "%d ", this->fullChain_(i).step);
    for(int ic=0;ic<this->chainDim_;ic++)
      fprintf(f_out, "%24.16lg ", this->fullChain_(i).state(ic));
    fprintf(f_out, "%24.16lg %24.16lg \n", this->fullChain_(i).alfa, this->fullChain_(i).post);
    
  }

  /// Close the text file
  if(fclose(f_out)){ 
    printf("writeChain: could not close file '%s'\n",filename.c_str()); 
    exit(1); 
  }

  /// Report
  printf("Written the states %d - %d to the text file %s\n", this->lastwrite_+1, 
         (int) this->fullChain_.XSize()-1, filename.c_str()); 

  return ;
  
}

void MCMC::writeFullChainTxt(string filename, Array1D<chainstate> fullchain)
{
 
  /// Open the text file in a write mode
  char* writemode=(char *)"w";

  /// Append if the names already prepended
  if (namePrepend_)
    writemode=(char *)"a";  

  FILE* f_out;
  if(!(f_out = fopen(filename.c_str(),writemode))){ 
    printf("writeChain: could not open file '%s'\n",filename.c_str()); 
    exit(1); 
  }

  /// Write the full chain
  for(int i=0;i<(int)fullchain.XSize();i++){
    fprintf(f_out, "%d ", fullchain(i).step);
    for(int ic=0;ic<this->chainDim_;ic++)
      fprintf(f_out, "%24.16lg ", fullchain(i).state(ic));
    fprintf(f_out, "%24.16lg %24.16lg \n", fullchain(i).alfa,fullchain(i).post);
    
  }

  /// Closte the file
 if(fclose(f_out)){ 
   printf("writeChain: could not close file '%s'\n",filename.c_str()); 
   exit(1); 
 }
 
 /// Report
 printf("Written the full chain out to text file %s\n",filename.c_str()); 

 return ;

}

void MCMC::writeChainBin(string filename)
{
 
  /// Choose whether write or append
  char* writemode=(char *)"wb";
  if (lastwrite_>=0)
    writemode=(char *)"ab";
  
  /// Open the binary file
  FILE* f_out;
  if(!(f_out = fopen(filename.c_str(),writemode))){ 
    printf("writeChain: could not open file '%s'\n",filename.c_str()); 
    exit(1); 
  }


  /// Write to the binary file
  for(int i=this->lastwrite_+1;i<(int) this->fullChain_.XSize();i++){
    fwrite(&(this->fullChain_(i).step), sizeof(int), 1, f_out);
    fwrite(fullChain_(i).state.GetArrayPointer(),this->chainDim_*sizeof(double),1, f_out);
    fwrite(&(this->fullChain_(i).alfa), sizeof(double), 1, f_out);
    fwrite(&(this->fullChain_(i).post), sizeof(double), 1, f_out);
  }

  /// CLose the binary file
  if(fclose(f_out)){ 
    printf("writeChain: could not close file '%s'\n",filename.c_str()); 
    exit(1); 
  }

  /// Report
  printf("Written the states %d - %d to the binary file %s\n", this->lastwrite_+1, 
         (int) this->fullChain_.XSize()-1, filename.c_str()); 
  
  return ;
  
}

void MCMC::parseBinChain(string filename, Array1D<chainstate>& readchain)
{
  double tmp;
  int readstep,i=0;

  FILE *fb = fopen(filename.c_str(),"rb");
  
  chainstate curchain;
  /// Read the binary file and write to an array of chain states
  while( fread(&readstep,sizeof(int),1,fb) ) {

    curchain.step=readstep;
    assert(readstep==i);
    
    curchain.state.Resize(this->chainDim_);
    
    fread(curchain.state.GetArrayPointer(),this->chainDim_*sizeof(double),1,fb);   
    fread(&tmp,sizeof(double),1,fb);
    curchain.alfa=tmp;
    fread(&tmp,sizeof(double),1,fb);
    curchain.post=tmp;
    readchain.PushBack(curchain);
    i++;  
  }
  
  fclose(fb);

  return;

}


void MCMC::getMode(Array1D<double>& MAPparams)
{
  
  //for(int ic=0;ic<this->chainDim_;ic++)
  MAPparams=modeState_.state;

  return;

}
