import numpy as npy

def pcorr(spl):
    """
    Compute Pearson's correlation coefficient between random vectors

    Args:
       spl - 2D array of samples, first dimensions is the number of samples,
            second dimension is the number of random vectors

    Output:
       Returns a 2D array of correlation coefficients between pairs of random vectors;
       only entries 0<=j<i<no. of random vectors are populated
       
    References:
       http://en.wikipedia.org/wiki/Pearson_product-moment_correlation_coefficient
     """
    nspl  = spl.shape[0]
    nvars = spl.shape[1]
    splmn  = npy.mean(spl,axis=0)
    splstd = npy.std(spl,axis=0)
    corrcoef = npy.zeros((nvars,nvars))
    for i in range(1,nvars):
        for j in range(i):
            rxy=npy.mean((spl[:,i]-splmn[i])*(spl[:,j]-splmn[j]))/(splstd[i]*splstd[j]);
            corrcoef[i,j] = rxy
    return corrcoef

def distcorr(spl):
    """
    Compute distance correlation between random vectors

    Args:
       spl - 2D array of samples, first dimensions is the number of samples,
             second dimension is the number of random vectors

    Output:
       Returns a 2D array of distance correlations between pairs of random vectors;
       only entries 0<=j<i<no. of random vectors are populated
       
    References:
       http://en.wikipedia.org/wiki/Distance_correlation
     """
    nspl  = spl.shape[0]
    nvars = spl.shape[1]
    if (nspl>5000):
        print 'Warning ! This might be a lengthy calculation: nspl=',nspl
    As=[]
    for i in range(nvars):
        print "Computing variable %d out of %d"%(i+1,nvars)
        Amat = npy.matrix(npy.zeros((nspl,nspl)))
        for i1 in range(nspl):
            for j1 in range(nspl):
                Amat[i1,j1] = abs(spl[i1,i]-spl[j1,i])
        # compute means
        Alin = npy.array([npy.mean(Amat[i1,:]) for i1 in range(nspl)])
        Acol = npy.array([npy.mean(Amat[:,j1]) for j1 in range(nspl)])
        Amn = npy.mean(Amat)
        # subtract/add means (linewise, columnwise, overall)
        Amat = Amat - Alin.reshape(nspl,1)
        Amat = (Amat.T - Acol.reshape(nspl,1)).T
        Amat = Amat+Amn
        As.append(Amat.copy())
    dCor = npy.zeros((nvars,nvars))
    dVarX = [npy.sqrt(npy.sum(npy.multiply(As[i],As[i]))/(nspl*nspl)) for i in range(nvars)]
    for i in range(1,nvars):
        for j in range(i):
            dCov   = npy.sqrt(npy.sum(npy.multiply(As[i],As[j]))/(nspl*nspl))
            dCor[i,j] = dCov/npy.sqrt(dVarX[i]*dVarX[j])
    As=[]
    return dCor

