#!/usr/bin/env python

import os, sys
import numpy as npy
import h5py
import matplotlib.pyplot as plt
from scipy import stats, mgrid, c_, reshape, random, rot90
from matplotlib.ticker import MultipleLocator, FormatStrFormatter


figname='cal_rip_pls_d12_wp.pdf'

vnames=["astem","br\_mr","br\_lit","br\_som","dr","gdd\_min","gdd\_max",
  "laimax","leaffall","lma","nue","q10\_mr","q10\_hr","rg\_frac","tleaf",
  "troot","tstem","tsmin","vc1","vc2","vc3","sc1","sc2"]

vpick=["q10\_mr","q10\_hr","br\_som","vc3","sc1","sc2"]

idx=[vnames.index(i) for i in vpick]

# Read 1D pdfs
x1D=[]
p1D=[]
for i in idx:
    file = h5py.File("dens_"+str(i+1)+'.h5', 'r')   
    x1D.append(npy.array(file["x"  ]))
    p1D.append(npy.array(file["p1D"]))
    file.close()

x2D = []
y2D = []
p2D = []
for j in range(1,len(idx)):
    for i in range(j):
        if ( idx[i] < idx[j] ):
            file = h5py.File("dens_"+str(idx[i]+1)+'_'+str(idx[j]+1)+'.h5', 'r')   
            prm = 0;
        else:
            file = h5py.File("dens_"+str(idx[j]+1)+'_'+str(idx[i]+1)+'.h5', 'r')   
            prm = 1;
        if (prm==1):
            x2D.append(npy.transpose(npy.array(file["y"  ])))
            y2D.append(npy.transpose(npy.array(file["x"  ])))
            p2D.append(npy.transpose(npy.array(file["p2D"])))
        else:
            x2D.append(npy.array(file["x"  ]))
            y2D.append(npy.array(file["y"  ]))
            p2D.append(npy.array(file["p2D"]))
        file.close()

x2D = npy.array(x2D)
y2D = npy.array(y2D)
p2D = npy.array(p2D)

nvars=len(x1D)

# Section 5
# ds is the distance between subplots
# xs,ys are the coordinates (normalized) of the subplot in the lower left corner
# xe,ye are the distances left in the uppper right corner
# ds are distances between subplots
# fsizex, fsizey are figure sizes
# ncont are no of contours for 2D pdfs
xs=0.05; ys=0.05; ds=0.005
xe=0.05; ye=0.05
fsizex=8; fsizey=8;
ncont=20; 
sx=(1-(nvars-1)*ds-xs-xe)/nvars;
sy=(1-(nvars-1)*ds-ys-ye)/nvars;

fs1=14

majorFormatter = FormatStrFormatter('%6.0e')

fig = plt.figure(figsize=(fsizex,fsizey))

# Section 5.1
subs=[]
# add diagonal plots
for i in range(nvars):
    subs.append(fig.add_axes([xs+i*(sx+ds),ys+(nvars-1-i)*(sy+ds),sx,sy]))

# add lower triangular plots
for i in range(nvars-1):
    for j in range(i+1):
        #print j,(nvars-2-i)
        subs.append(fig.add_axes([xs+j*(sx+ds),ys+(nvars-2-i)*(sy+ds),sx,sy]))

subsnp=npy.array(subs)

# Plot 1D pdfs
for i in range(nvars):
    pl1d,=subsnp[i].plot(x1D[i],p1D[i])

# Plot 2D pdfs
for i in range(nvars*(nvars-1)/2):
    pl2d=subsnp[nvars+i].contour(x2D[i],y2D[i],p2D[i],ncont)

# Section 5.2
# no ticks and ticklabels
for i in range(nvars,len(subsnp)):
    xti=subsnp[i].set_xticks([]); 
    yti=subsnp[i].set_yticks([]); 

# for diagonal plots only put xmin and xmax
#for i in range(1,nvars+1):
#    subsnp[i-1].set_xticks([d0[nskip:,i].min(),d0[nskip:,i].max()]); 

# Set y labels on the right for diagonal plots
for i in range(nvars):
    subsnp[i].xaxis.tick_top()
    subsnp[i].yaxis.tick_right()
    subsnp[i].yaxis.set_label_position("right")
    #subsnp[i].set_ylabel(r'$'+str(i+1)+'$', fontsize=fs1, rotation=0)

icnt=0
for i in range(len(subsnp)-nvars+1,len(subsnp)):
    icnt+=1
    subsnp[i].set_xlabel(r'$'+vpick[icnt-1]+'$', fontsize=fs1)

icnt=nvars
for i in range(nvars,2*nvars-1):
    icnt+=i-nvars
    subsnp[icnt].set_ylabel(r'$'+vpick[i-nvars+1]+'$', fontsize=fs1, rotation=90)

subsnp[0].yaxis.set_label_position("left")
subsnp[0].set_ylabel(r'$'+vpick[0]+'$', fontsize=fs1, rotation=90)
subsnp[nvars-1].set_xlabel(r'$'+vpick[nvars-1]+'$', fontsize=fs1)

# Custom labels
subsnp[0].set_xticks([1,1.1,1.2,1.3]);
subsnp[0].set_yticks([4,8,12]);
subsnp[1].set_xticks([1.5,2.25,3]);
subsnp[1].set_yticks([1, 2, 3]); 
subsnp[2].set_xlim([0,0.00033])	 
subsnp[2].set_xticks([0.0001,0.0003]);
subsnp[2].set_yticks([10000,20000]);
subsnp[2].set_yticklabels(["$10^{4}$","$2\cdot 10^{4}$"],fontsize=fs1-2)
subsnp[3].set_xlim([0,1300])	 
subsnp[3].set_xticks([500,1000]);
subsnp[3].set_yticks([0.001,0.002]);
subsnp[4].set_xlim([0,300])	 
subsnp[4].set_xticks([100,200]);
subsnp[4].set_yticks([0.02,0.04]);
subsnp[5].set_xlim([100,14000])	 
subsnp[5].set_xticks([4000,10000]);
subsnp[5].set_yticks([0.0001,0.0002]);
subsnp[5].set_yticklabels(["$10^{-4}$","$2\cdot 10^{-4}$"],fontsize=fs1-2,rotation=60)

#plt.show()

plt.savefig(figname)

